@@ -1,4 +1,5 @@
 from deezer.client import Client
+from deezer.pagination import PaginatedList
 from deezer.resources import (
     Album,
     Artist,
@@ -10,9 +11,10 @@ from deezer.resources import (
     User,
 )
 
-__version__ = "4.2.1"
+__version__ = "5.0.0"
 __all__ = [
     "Client",
+    "PaginatedList",
     "Resource",
     "Album",
     "Artist",
