@@ -2,7 +2,9 @@
 Implements a client class to query the
 `Deezer API <https://developers.deezer.com/api>`_
 """
-from typing import Any, Dict, List, Optional, Type, Union
+from __future__ import annotations
+
+from typing import Any
 
 import requests
 
@@ -11,6 +13,7 @@ from deezer.exceptions import (
     DeezerHTTPError,
     DeezerUnknownResource,
 )
+from deezer.pagination import PaginatedList
 from deezer.resources import (
     Album,
     Artist,
@@ -76,15 +79,15 @@ class Client:
         self.access_token = access_token
         self.session = requests.Session()
 
-        # Headers
-        if headers:
-            self.session.headers.update(headers)
+        headers = headers or {}
+        self.session.headers.update(headers)
 
     def _process_json(
         self,
-        item: Dict[str, Any],
-        parent: Optional[Resource] = None,
-        resource_type: Optional[Type[Resource]] = None,
+        item: dict[str, Any],
+        parent: Resource | None = None,
+        resource_type: type[Resource] | None = None,
+        paginate_list=False,
     ):
         """
         Recursively convert dictionary
@@ -93,10 +96,17 @@ class Client:
         :param item: the JSON response as dict.
         :param parent: A reference to the parent resource, to avoid fetching again.
         :param resource_type: The resource class to use as top level.
+        :param paginate_list: Whether to wrap list into a pagination object.
         :returns: instance of :class:`~deezer.resources.Resource`
         """
         if "data" in item:
-            return [self._process_json(i, parent) for i in item["data"]]
+            parsed_data = [
+                self._process_json(i, parent, paginate_list=False) for i in item["data"]
+            ]
+            if not paginate_list:
+                return parsed_data
+            item["data"] = parsed_data
+            return item
 
         result = {}
         for key, value in item.items():
@@ -122,8 +132,9 @@ class Client:
         self,
         method: str,
         path: str,
-        parent: Optional[Resource] = None,
-        resource_type: Optional[Type[Resource]] = None,
+        parent: Resource | None = None,
+        resource_type: type[Resource] | None = None,
+        paginate_list=False,
         **params,
     ):
         """
@@ -133,7 +144,8 @@ class Client:
         :param path: The path to make the API call to (e.g. 'artist/1234').
         :param parent: A reference to the parent resource, to avoid fetching again.
         :param resource_type: The resource class to use as top level.
-        :param params: Query parameters to add the the request
+        :param paginate_list: Whether to wrap list into a pagination object.
+        :param params: Query parameters to add to the request
         """
         if self.access_token is not None:
             params["access_token"] = str(self.access_token)
@@ -151,7 +163,15 @@ class Client:
             return json_data
         if "error" in json_data:
             raise DeezerErrorResponse(json_data)
-        return self._process_json(json_data, parent=parent, resource_type=resource_type)
+        return self._process_json(
+            json_data,
+            parent=parent,
+            resource_type=resource_type,
+            paginate_list=paginate_list,
+        )
+
+    def _get_paginated_list(self, path, **params):
+        return PaginatedList(client=self, base_path=path, **params)
 
     def get_album(self, album_id: int) -> Album:
         """
@@ -183,13 +203,13 @@ class Client:
         """
         Get overall charts for tracks, albums, artists and playlists.
 
-        Combines charts of several resources in one endpoint.
+        Combine charts of several resources in one endpoint.
 
         :returns: a :class:`~deezer.resources.Chart` instance.
         """
         return self.request("GET", "chart", resource_type=Chart)
 
-    def get_tracks_chart(self) -> List[Track]:
+    def get_tracks_chart(self) -> list[Track]:
         """
         Get top tracks.
 
@@ -197,7 +217,7 @@ class Client:
         """
         return self.request("GET", "chart/0/tracks")
 
-    def get_albums_chart(self) -> List[Album]:
+    def get_albums_chart(self) -> list[Album]:
         """
         Get top albums.
 
@@ -205,7 +225,7 @@ class Client:
         """
         return self.request("GET", "chart/0/albums")
 
-    def get_artists_chart(self) -> List[Artist]:
+    def get_artists_chart(self) -> list[Artist]:
         """
         Get top artists.
 
@@ -213,7 +233,7 @@ class Client:
         """
         return self.request("GET", "chart/0/artists")
 
-    def get_playlists_chart(self) -> List[Playlist]:
+    def get_playlists_chart(self) -> list[Playlist]:
         """
         Get top playlists.
 
@@ -221,7 +241,7 @@ class Client:
         """
         return self.request("GET", "chart/0/playlists")
 
-    def get_podcasts_chart(self) -> List[Podcast]:
+    def get_podcasts_chart(self) -> list[Podcast]:
         """
         Get top podcasts.
 
@@ -245,13 +265,14 @@ class Client:
         """
         return self.request("GET", f"genre/{genre_id}")
 
-    def list_genres(self) -> List[Genre]:
+    def list_genres(self) -> PaginatedList[Genre]:
         """
         List musical genres.
 
-        :returns: a list of :class:`~deezer.resources.Genre` objects.
+        :returns: a :class:`~deezer.pagination.PaginatedList`
+                  of :class:`~deezer.resources.Genre` objects.
         """
-        return self.request("GET", "genre")
+        return self._get_paginated_list("genre")
 
     def get_playlist(self, playlist_id: int) -> Playlist:
         """
@@ -271,27 +292,29 @@ class Client:
 
     def get_radio(self, radio_id: int) -> Radio:
         """
-        Get the radio with the given ID..
+        Get the radio with the given ID.
 
         :returns: a :class:`~deezer.resources.Radio` object
         """
         return self.request("GET", f"radio/{radio_id}")
 
-    def list_radios(self) -> List[Radio]:
+    def list_radios(self) -> PaginatedList[Radio]:
         """
         List radios.
 
-        :returns: a list of :class:`~deezer.resources.Radio` objects
+        :returns: a :class:`~deezer.pagination.PaginatedList`
+                  of :class:`~deezer.resources.Radio` objects.
         """
-        return self.request("GET", "radio")
+        return self._get_paginated_list("radio")
 
-    def get_radios_top(self):
+    def get_radios_top(self) -> PaginatedList[Radio]:
         """
-        Get the top radios (5 radios).
+        Get the top radios.
 
-        :returns: a :class:`~deezer.resources.Radio` object
+        :returns: a :class:`~deezer.pagination.PaginatedList`
+                  of :class:`~deezer.resources.Radio` objects.
         """
-        return self.request("GET", "radio/top")
+        return self._get_paginated_list("radio/top")
 
     def get_track(self, track_id: int) -> Track:
         """
@@ -301,7 +324,7 @@ class Client:
         """
         return self.request("GET", f"track/{track_id}")
 
-    def get_user(self, user_id: Optional[int] = None) -> User:
+    def get_user(self, user_id: int | None = None) -> User:
         """
         Get the user with the given ID.
 
@@ -310,7 +333,7 @@ class Client:
         user_id_str = str(user_id) if user_id else "me"
         return self.request("GET", f"user/{user_id_str}")
 
-    def get_user_albums(self, user_id: Optional[int] = None) -> List[Album]:
+    def get_user_albums(self, user_id: int | None = None) -> PaginatedList:
         """
         Get the favourites albums for the given user_id if provided or current user if not.
 
@@ -318,7 +341,7 @@ class Client:
         :return: a list of :class:`~deezer.resources.Album` instances.
         """
         user_id_str = str(user_id) if user_id else "me"
-        return self.request("GET", f"user/{user_id_str}/albums")
+        return self._get_paginated_list(f"user/{user_id_str}/albums")
 
     def add_user_album(self, album_id: int) -> bool:
         """
@@ -338,15 +361,16 @@ class Client:
         """
         return self.request("DELETE", "user/me/albums", album_id=album_id)
 
-    def get_user_artists(self, user_id: Optional[int] = None) -> List[Artist]:
+    def get_user_artists(self, user_id: int | None = None) -> PaginatedList[Artist]:
         """
         Get the favourites artists for the given user_id if provided or current user if not.
 
         :param user_id: the user ID to get favourites artists.
-        :return: a list of :class:`~deezer.resources.Artist` instances.
+        :return: a :class:`~deezer.pagination.PaginatedList`
+                 of :class:`~deezer.resources.Artist` instances.
         """
         user_id_str = str(user_id) if user_id else "me"
-        return self.request("GET", f"user/{user_id_str}/artists")
+        return self._get_paginated_list(f"user/{user_id_str}/artists")
 
     def add_user_artist(self, artist_id: int) -> bool:
         """
@@ -366,23 +390,25 @@ class Client:
         """
         return self.request("DELETE", "user/me/artists", artist_id=artist_id)
 
-    def get_user_history(self) -> List[Track]:
+    def get_user_history(self) -> PaginatedList[Track]:
         """
         Returns a list of the recently played tracks for the current user.
 
-        :return: a list of :class:`~deezer.resources.Track` instances.
+        :return: a :class:`~deezer.pagination.PaginatedList`
+                 of :class:`~deezer.resources.Track` instances.
         """
-        return self.request("GET", "user/me/history")
+        return self._get_paginated_list("user/me/history")
 
-    def get_user_tracks(self, user_id: Optional[int] = None) -> List[Track]:
+    def get_user_tracks(self, user_id: int | None = None) -> PaginatedList[Track]:
         """
         Get the favourites tracks for the given user_id if provided or current user if not.
 
         :param user_id: the user ID to get favourites tracks.
-        :return: a list of :class:`~deezer.resources.Track` instances.
+        :return: a :class:`~deezer.pagination.PaginatedList`
+                 of :class:`~deezer.resources.Track` instances.
         """
         user_id_str = str(user_id) if user_id else "me"
-        return self.request("GET", f"user/{user_id_str}/tracks")
+        return self._get_paginated_list(f"user/{user_id_str}/tracks")
 
     def add_user_track(self, track_id: int) -> bool:
         """
@@ -406,30 +432,23 @@ class Client:
         self,
         path: str,
         query: str = "",
-        strict: Optional[bool] = None,
-        ordering: Optional[str] = None,
-        index: Optional[int] = None,
-        limit: Optional[int] = None,
-        **advanced_params: Optional[Union[str, int]],
+        strict: bool | None = None,
+        ordering: str | None = None,
+        **advanced_params: str | int | None,
     ):
         optional_params = {}
         if strict is True:
             optional_params["strict"] = "on"
         if ordering:
             optional_params["ordering"] = ordering
-        if index:
-            optional_params["index"] = index
-        if limit:
-            optional_params["limit"] = limit
         query_parts = []
         if query:
             query_parts.append(query)
         for param_name, param_value in advanced_params.items():
             if param_value:
                 query_parts.append(f'{param_name}:"{param_value}"')
-        return self.request(
-            "GET",
-            f"search/{path}" if path else "search",
+        return self._get_paginated_list(
+            path=f"search/{path}" if path else "search",
             q=" ".join(query_parts),
             **optional_params,
         )
@@ -437,18 +456,16 @@ class Client:
     def search(
         self,
         query: str = "",
-        strict: Optional[bool] = None,
-        ordering: Optional[str] = None,
-        artist: Optional[str] = None,
-        album: Optional[str] = None,
-        track: Optional[str] = None,
-        label: Optional[str] = None,
-        dur_min: Optional[int] = None,
-        dur_max: Optional[int] = None,
-        bpm_min: Optional[int] = None,
-        bpm_max: Optional[int] = None,
-        index: Optional[int] = None,
-        limit: Optional[int] = None,
+        strict: bool | None = None,
+        ordering: str | None = None,
+        artist: str | None = None,
+        album: str | None = None,
+        track: str | None = None,
+        label: str | None = None,
+        dur_min: int | None = None,
+        dur_max: int | None = None,
+        bpm_min: int | None = None,
+        bpm_max: int | None = None,
     ):
         """
         Search tracks.
@@ -458,7 +475,7 @@ class Client:
 
         :param query: the query to search for, this is directly passed as q query.
         :param strict: whether to disable fuzzy search and enable strict mode.
-        :param ordering: see Deezer's API docs for possible values.
+        :param ordering: see Deezer API docs for possible values.
         :param artist: parameter for the advanced search feature.
         :param album: parameter for the advanced search feature.
         :param track: parameter for the advanced search feature.
@@ -467,8 +484,6 @@ class Client:
         :param dur_max: parameter for the advanced search feature.
         :param bpm_min: parameter for the advanced search feature.
         :param bpm_max: parameter for the advanced search feature.
-        :param index: the offset of the first object you want to get.
-        :param limit: the maximum number of objects to return.
         :returns: a list of :class:`~deezer.resources.Track` instances.
         """
         return self._search(
@@ -484,26 +499,20 @@ class Client:
             dur_max=dur_max,
             bpm_min=bpm_min,
             bpm_max=bpm_max,
-            index=index,
-            limit=limit,
         )
 
     def search_albums(
         self,
         query: str = "",
-        strict: Optional[bool] = None,
-        ordering: Optional[str] = None,
-        index: Optional[int] = None,
-        limit: Optional[int] = None,
-    ) -> List[Album]:
+        strict: bool | None = None,
+        ordering: str | None = None,
+    ) -> PaginatedList[Album]:
         """
         Search albums matching the given query.
 
         :param query: the query to search for, this is directly passed as q query.
         :param strict: whether to disable fuzzy search and enable strict mode.
-        :param ordering: see Deezer's API docs for possible values.
-        :param index: the offset of the first object you want to get.
-        :param limit: the maximum number of objects to return.
+        :param ordering: see Deezer API docs for possible values.
         :return: list of :class:`~deezer.resources.Album` instances.
         """
         return self._search(
@@ -511,26 +520,20 @@ class Client:
             query=query,
             strict=strict,
             ordering=ordering,
-            index=index,
-            limit=limit,
         )
 
     def search_artists(
         self,
         query: str = "",
-        strict: Optional[bool] = None,
-        ordering: Optional[str] = None,
-        index: Optional[int] = None,
-        limit: Optional[int] = None,
-    ) -> List[Artist]:
+        strict: bool | None = None,
+        ordering: str | None = None,
+    ) -> PaginatedList[Artist]:
         """
         Search artists matching the given query.
 
         :param query: the query to search for, this is directly passed as q query.
         :param strict: whether to disable fuzzy search and enable strict mode.
-        :param ordering: see Deezer's API docs for possible values.
-        :param index: the offset of the first object you want to get.
-        :param limit: the maximum number of objects to return.
+        :param ordering: see Deezer API docs for possible values.
         :return: list of :class:`~deezer.resources.Album` instances.
         """
         return self._search(
@@ -538,6 +541,4 @@ class Client:
             query=query,
             strict=strict,
             ordering=ordering,
-            index=index,
-            limit=limit,
         )
